@@ -9,6 +9,14 @@
 set -o xtrace
 set -o errexit
 
+on_exit () {
+  if [ -e ./unstable-chroot/debootstrap/debootstrap.log ]; then
+    echo "Dumping debootstrap.log"
+    cat ./unstable-chroot/debootstrap/debootstrap.log
+  fi
+}
+trap on_exit EXIT
+
 if [ "${IS_PATCH}" = "true" ]; then
   git diff HEAD -- . ':!debian' > ../upstream.patch
   git diff HEAD -- debian > ../debian.patch
@@ -32,12 +40,11 @@ fi
 cd ..
 
 git clone https://salsa.debian.org/installer-team/debootstrap.git debootstrap.git
-mkdir unstable-chroot
 export DEBOOTSTRAP_DIR=`pwd`/debootstrap.git
-# perl-openssl-defaults is explicitly listed to work around https://bugs.debian.org/907015
-sudo -E ./debootstrap.git/debootstrap --include=build-essential,perl-openssl-defaults,git-buildpackage,fakeroot,debhelper,cmake,libssl-dev,pkg-config,python3-sphinx,zlib1g-dev,libicu-dev,libsasl2-dev,libsnappy-dev,libzstd-dev unstable ./unstable-chroot/ http://cdn-aws.deb.debian.org/debian
+sudo -E ./debootstrap.git/debootstrap unstable ./unstable-chroot/ http://cdn-aws.deb.debian.org/debian
 cp -a mongoc ./unstable-chroot/tmp/
 sudo chroot ./unstable-chroot /bin/bash -c "(set -o xtrace && \
+  apt-get install -y build-essential git-buildpackage fakeroot debhelper cmake libssl-dev pkg-config python3-sphinx zlib1g-dev libicu-dev libsasl2-dev libsnappy-dev libzstd-dev && \
   cd /tmp/mongoc && \
   git clean -fdx && \
   git reset --hard HEAD && \
