@@ -99,7 +99,7 @@ test_rs_failover (void)
    request =
       mock_rs_receives_msg (rs, 0, tmp_bson ("{'insert': 'collection'}"), b);
    BSON_ASSERT (mock_rs_request_is_to_secondary (rs, request));
-   mock_server_replies_simple (request, "{'ok': 0, 'errmsg': 'not master'}");
+   mock_server_replies_simple (request, "{'ok': 0, 'code': 10107, 'errmsg': 'not master'}");
    request_destroy (request);
 
    request =
@@ -147,7 +147,7 @@ test_command_with_opts (void *ctx)
    collection = get_test_collection (client, "retryable_writes");
 
    if (!mongoc_collection_drop (collection, &error)) {
-      if (strcmp (error.message, "ns not found")) {
+      if (NULL == strstr (error.message, "ns not found")) {
          /* an error besides ns not found */
          ASSERT_OR_PRINT (false, error);
       }
